#!/usr/bin/python3

import atexit
import config
import loghandler
import netifaces
import os
import socket
import subprocess
import sys
import threading
import time
import uuid

def cond_unlink(path, log):
    try:
        os.unlink(path)
        log.DEBUG('removed %s' % path)
    except FileNotFoundError:
        pass

class AddrInfo:

    def __init__(self, host):
        self.addr = [ h[4][0] for h in
                      socket.getaddrinfo(host, 22,
                                         type=socket.SocketKind.SOCK_STREAM) ]
        names = set([ socket.gethostbyaddr(a)[0] for a in self.addr ])
        if len(names) != 1:
            raise Exception('%s resolves to more than one name',
                            self.addr, names)
        self.name = names.pop()

    def __repr__(self):
        return 'AddrInfo(%s, %s)' % (self.name, self.addr)
    
class Server:

    def __init__(self, options, config, entry, path, uuid, log):
        self.options = options
        self.config = config
        self.entry = entry
        self.path = path
        self.log = log
        self.uuid = uuid
        self.socket_path = '/tmp/%s_server' % (self.uuid)
        self.mutex = threading.Lock()
        self.thread_md5 = None
        self.thread_cpio = None
        self.thread_server = threading.Thread(daemon=True, target=self.run)
        self.thread_server.start()
        
    def run(self):
        atexit.register(cond_unlink, self.socket_path, self.log)
        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(self.socket_path)
        server.listen(2)
        config_MD5_socket,_ = server.accept()
        self.log.DEBUG('config+MD5', config_MD5_socket)
        c = config_MD5_socket.makefile('r').read()
        if config.parse(c).asList() != self.config.asList():
            raise Exception('Configuration differs')
        with self.mutex:
            self.thread_md5 = threading.Thread(daemon=True,
                                               target=self.send_MD5,
                                               args=(config_MD5_socket,))
            self.thread_md5.start()

        cpio_socket,_ = server.accept()
        self.log.DEBUG('CPIO', cpio_socket)
        with self.mutex:
            self.thread_md5 = threading.Thread(daemon=True,
                                               target=self.run_cpio,
                                               args=(cpio_socket,))
            self.thread_md5.start()
        
        cond_unlink(self.socket_path, self.log)

    def pending(self):
        with self.mutex:
            return ((self.thread_md5 and self.thread_md5.thread.is_alive()) or
                    (self.thread_cpio and self.thread_cpio.thread.is_alive()) or
                    (self.thread_server.is_alive()))
        
    def send_MD5(self, config_MD5):
        self.log.DEBUG('send_MD5',
                       self.config.primary.mount.path,
                       self.path)
        cmd = (
            [ '/usr/bin/md5toc' ] +
            ( self.options.xattr and
              [ '--xattr' ] or []) +
            ( self.options.xattr and self.options.max_age and
              [ '--max-age', self.options.max_age ] or []) +
            [ '.' ]
        )
        cwd = os.path.join(self.config.primary.mount.path, self.path)
        stdout = config_MD5.makefile('wb')
        try:
            subprocess.check_call(cmd, cwd=cwd, stdout=stdout)
        finally:
            config_MD5.shutdown(socket.SHUT_RDWR)
            config_MD5.close()
            
    def run_cpio(self, cpio_socket):
        self.log.DEBUG('START run_cpio',
                       self.config.primary.mount.path,
                       self.path)
        cmd = [ '/bin/cpio', '-oc', '--quiet' ]
        cwd = os.path.join(self.config.primary.mount.path, self.path)
        stdin =  cpio_socket.makefile('rb')
        stdout =  cpio_socket.makefile('wb')
        try:
            subprocess.check_call(cmd, cwd=cwd,
                                  stdin=stdin,
                                  stdout=stdout,
                                  stderr=self.log.makefile(encoding='utf-8'))
            self.log.DEBUG('OK run_cpio', cmd, cwd)
        finally:
            cpio_socket.shutdown(socket.SHUT_RDWR)
        self.log.DEBUG('DONE run_cpio', cmd, cwd)

class Client:

    def __init__(self, options, config, entry, log):
        self.options = options
        self.config = config
        self.entry = entry
        self.uuid = str(uuid.uuid4())
        self.log = loghandler.LOG(parent=log,
                                  prefix="%s client " % (self.uuid))
        self.server_log = loghandler.LOG(parent=log,
                                         prefix="%s server " % (self.uuid))
        self.thread = threading.Thread(daemon=True, target=self.run)
        self.thread.start()
        
    def pending(self):
        return self.thread.is_alive()
    
    def run(self):
        self.log.DEBUG('Running', self.entry.dump())
        for path in self.entry.path:
            readable = '%s:%s/%s -> %s:%s/%s' % (
                self.config.primary.mount.host,
                self.config.primary.mount.path,
                path,
                self.entry.mount.host,
                self.entry.mount.path,
                path)
            self.log.MESSAGE('START %s' % (readable))
            server = Server(options=self.options,
                            config=self.config,
                            entry=self.entry,
                            path=path,
                            uuid=self.uuid,
                            log=self.server_log)
            socket_path = '/tmp/%s_client' % (self.uuid)
            cmd = (
                [ 'ssh', self.entry.mount.host ] +
                ( self.options.user and
                  [ '-l', self.options.user ] or []) +
                ( self.options.identity and
                  [ '-i', self.options.identity ] or []) +
                [ '-R', '%s:%s' % (socket_path, server.socket_path) ] +
                [ os.path.realpath(sys.argv[0]), 
                '--secondary', socket_path, self.entry.mount.path, path ] +
                ( self.options.debug and
                  [ '--debug' ] or []) +
                ( self.options.xattr and
                  [ '--xattr' ] or []) +
                ( self.options.xattr and self.options.max_age and
                  [ '--max-age', self.options.max_age ] or []) 
            )
            self.log.DEBUG('CMD="%s"' % (' '.join(cmd)))
            stdout = loghandler.LOG(parent=self.log, prefix='STDOUT ')
            stderr = loghandler.LOG(parent=self.log, prefix='STDERR ')
            subprocess.check_call(cmd,
                                  stdout=stdout.makefile(encoding='utf-8'),
                                  stderr=stderr.makefile(encoding='utf-8'))
            time.sleep(1)
            self.log.MESSAGE('DONE %s' % (readable))
            time.sleep(1)
   
def do_backup(options, config):
    def is_primary():
        node = set()
        for i in netifaces.interfaces():
            for k,al in netifaces.ifaddresses(i).items():
                if k in [netifaces.AF_INET, netifaces.AF_INET6]:
                    for a in al:
                        node.add(a['addr'])
        primary = AddrInfo(config.primary.mount.host)
        return any(node.intersection(primary.addr))

    if not is_primary():
        raise Exception('Host is not primary')

    if options.debug:
        log = loghandler.LOG(loghandler.LOG_DEBUG)
    else:
        log = loghandler.LOG(loghandler.LOG_WARNING)
    
    for b in config.secondary.backup:
        client = [ Client(options=options,
                          config=config,
                          entry=e,
                          log=log) for e in b.entry ]
        while any([ c.pending() for c in client ]):
            time.sleep(1)


